from fixtures import MonkeyPatch


class Overlay(MonkeyPatch):

    def __init__(self, name, overlay, condition):
        super(Overlay, self).__init__(name, self._new_value)
        self.overlay = overlay
        self.condition = condition

    def _setUp(self):
        # XXX copied from MonkeyPatch
        location, attribute = self.name.rsplit('.', 1)
        __import__(location, {}, {})
        components = location.split('.')
        current = __import__(components[0], {}, {})
        for component in components[1:]:
            current = getattr(current, component)
        self._old_value = getattr(current, attribute)
        super(Overlay, self)._setUp()

    def _new_value(self, *args, **kwargs):
        if self.condition(*args, **kwargs):
            return self.overlay(self._old_value, *args, **kwargs)
        else:
            return self._old_value(*args, **kwargs)



import os

from testtools.matchers import (
    AfterPreprocessing,
    Equals,
)


def HasOwnership(uid, gid):

    def ownership(path):
        info = os.stat(path)
        return info.st_uid, info.st_gid

    return AfterPreprocessing(ownership, Equals((uid, gid)))



import pwd

from testtools import TestCase

from ..users import FakeUsers


class FakeUsersTest(TestCase):

    def setUp(self):
        super(FakeUsersTest, self).setUp()
        self.users = self.useFixture(FakeUsers())

    def test_real(self):
        info = pwd.getpwnam("root")
        self.assertEqual(0, info.pw_uid)

    def test_fake(self):
        self.users.add("foo", 123)
        info = pwd.getpwnam("foo")
        self.assertEqual(123, info.pw_uid)



import time

from testtools import TestCase

from ..time import FakeTime


class FakeTimeTest(TestCase):

    def setUp(self):
        super(FakeTimeTest, self).setUp()
        self.time = self.useFixture(FakeTime())

    def test_time(self):
        stamp1 = time.time()
        stamp2 = time.time()
        self.assertEqual(stamp2, stamp1)

    def test_sleep(self):
        stamp1 = time.time()
        time.sleep(1)
        stamp2 = time.time()
        self.assertEqual(1, stamp2 - stamp1)

    def test_reset(self):
        self.time.set(123)
        self.assertEqual(123, time.time())



import os
from testtools import TestCase
from testtools.matchers import FileContains

from fixtures import TempDir

from ..wget import Wget


class WgetTest(TestCase):

    def setUp(self):
        super(WgetTest, self).setUp()
        self.locations = {"http://x": b"data"}
        self.wget = Wget(locations=self.locations)

    def test_to_stdout(self):
        result = self.wget({"args": ["wget","-N", "-O", "-", "http://x"]})
        self.assertEqual(b"data", result["stdout"].getvalue())

    def test_to_file(self):
        temp_dir = self.useFixture(TempDir())
        path = temp_dir.join("output")
        self.wget({"args": ["wget", "-N", "-O", path, "http://x"]})
        self.assertThat(path, FileContains("data"))

    def test_to_default(self):
        self.wget({"args": ["wget", "-N", "http://x"]})
        self.assertThat("x", FileContains("data"))



from fixtures import Fixture

from fakesleep import (
    monkey_patch,
    monkey_restore,
    reset,
)


class FakeTime(Fixture):
    """Fixture adapter around fakesleep."""

    def set(self, seconds=None):
        """Set the global fake time to the given epoch or the real current one.

        :param seconds: Fake current time, in seconds since the epoch. If
             ``None``, use the real current time.
        """
        reset(seconds=seconds)

    def _setUp(self):
        monkey_patch()
        self.addCleanup(monkey_restore)



from testtools import TestCase

from ..dpkg import Dpkg


class DpkgTest(TestCase):

    def setUp(self):
        super(DpkgTest, self).setUp()
        self.dpkg = Dpkg()

    def test_install(self):
        self.dpkg({"args": ["dpkg", "-i", "foo_1.0-1.deb"]})
        self.assertEqual(["install"], self.dpkg.actions["foo"])



from fixtures import MonkeyPatch


class FakeThreads(MonkeyPatch):

    def __init__(self):
        super(FakeThreads, self).__init__("threading.Thread", self)
        self._threads = []
        self._hang = False

    def hang(self, flag=True):
        self._hang = flag

    def __getitem__(self, index):
        return self._threads[index]

    def __call__(self, *args, **kwargs):
        thread = _Thread(*args, **kwargs)
        if not thread.name:
            thread.name = "fake-thread-{}".format(len(self._threads))
        thread.hang = self._hang
        self._threads.append(thread)
        return thread


class _Thread(object):

    def __init__(self, group=None, target=None, name=None, args=(),
                 kwargs={}, daemon=None):
        self.group = group
        self.target = target
        self.name = name
        self.args = args
        self.kwargs = kwargs
        self.daemon = daemon
        self.alive = None  # None -> not started, True/False -> started stopped
        self.hang = None

    def start(self):
        if self.alive is not None:
            raise RuntimeError("threads can only be started once")
        self.alive = True
        if not self.hang:
            self.target(*self.args, **self.kwargs)
            self.alive = False

    def join(self, timeout=None):
        if self.alive is None:
            # It's an an error to join() a thread before it has been started
            raise RuntimeError("cannot join thread before it is started")
        if self.hang:
            if timeout is None:
                raise AssertionError(
                    "can't simulate hung thread with no timeout")

    def isAlive(self):
        return bool(self.alive)

    is_alive = isAlive



import io
import subprocess

from testtools import TestCase

from ..fixture import FakeProcesses


class FakeProcessesTest(TestCase):

    def setUp(self):
        super(FakeProcessesTest, self).setUp()
        self.processes = self.useFixture(FakeProcesses())

    def test_real(self):
        self.assertEqual(b"hi\n", subprocess.check_output(["echo", "hi"]))

    def test_fake(self):

        def get_info(proc_args):
            return {"stdout": io.BytesIO(b"hi!")}

        self.processes.add(get_info, name="echo")
        self.assertIs(get_info, self.processes.echo)
        self.assertEqual(b"hi!", subprocess.check_output(["echo", "hi"]))



import threading

from testtools import TestCase

from ..threads import FakeThreads


class FakeThreadsTest(TestCase):

    def setUp(self):
        super(FakeThreadsTest, self).setUp()
        self.threads = self.useFixture(FakeThreads())

    def test_getitem(self):
        thread = threading.Thread()
        self.assertIs(thread, self.threads[0])

    def test_start(self):
        calls = [None]
        thread = threading.Thread(target=calls.pop)
        self.assertFalse(thread.isAlive())
        thread.start()
        self.assertEqual([], calls)
        self.assertFalse(thread.isAlive())

    def test_twice(self):
        thread = threading.Thread(target=lambda: None)
        thread.start()
        error = self.assertRaises(RuntimeError, thread.start)
        self.assertEqual("threads can only be started once", str(error))

    def test_join_before_starting(self):
        thread = threading.Thread()
        error = self.assertRaises(RuntimeError, thread.join)
        self.assertEqual("cannot join thread before it is started", str(error))

    def test_hang(self):
        self.threads.hang()
        thread = threading.Thread()
        thread.start()
        thread.join(timeout=1)
        self.assertTrue(thread.isAlive())

    def test_hang_without_timeout(self):
        self.threads.hang()
        thread = threading.Thread()
        thread.start()
        error = self.assertRaises(AssertionError, thread.join)
        self.assertEqual(
            "can't simulate hung thread with no timeout", str(error))



import io


class Systemctl(object):

    name = "systemctl"

    def __init__(self):
        self.actions = {}

    def __call__(self, proc_args):
        action, service_name = proc_args["args"][1:]
        actions = self.actions.setdefault(service_name, [])
        returncode = 0
        stdout = io.BytesIO()
        if action == "is-active":
            if not actions or actions[-1] != "start":
                returncode = 3
                stdout.write(b"inactive\n")
            else:
                stdout.write(b"active\n")
        else:
            actions.append(action)
        return {"stdout": stdout, "returncode": returncode}



import six
import subprocess

from fixtures import FakePopen
from fixtures._fixtures import popen


class FakeProcesses(FakePopen):
    """Enhances FakePopen by supporting multiple processes.

    This is essentially a registry to dispatch calls to FakePopen to
    fake code associated with an executable name, or to fall back to
    real executables in a matching name is not registered.
    """

    def __init__(self):
        self._registry = {}
        self._real_Popen = subprocess.Popen

    def add(self, process, name=None):
        """Add a new process to the registry.

        :param process: A callable (either plain function or object
            implementing __calll).
        :param name: The name of the executable to match. If not given
            it must be provided as 'name' attribute of the given `process`.
            callable.
        """
        name = name or process.name
        assert name, "No executable name given."""
        self._registry[name] = process

    def get_info(self, proc_args):
        name = proc_args["args"][0]
        get_info = self._registry[name]
        return get_info(proc_args)

    def __call__(self, *args, **kwargs):
        name = args[0][0]
        if name not in self._registry:
            return self._real_Popen(*args, **kwargs)
        return super(FakeProcesses, self).__call__(*args, **kwargs)

    def __getattr__(self, name):
        return self._registry[name]


class _FakeProcessWithMissingAPIs(popen.FakeProcess):
    """Can be dropped once:

    https://github.com/testing-cabal/fixtures/pull/32.

    is merged upstream.
    """

    if six.PY3:

        @property
        def args(self):
            return self._args["args"]

    def poll(self):
        return self.returncode

    def communicate(self, input=None, timeout=None):
        return super(_FakeProcessWithMissingAPIs, self).communicate()


popen.FakeProcess = _FakeProcessWithMissingAPIs



import requests

from testtools import TestCase

from ..network import FakeNetwork


class FakeGroupsTest(TestCase):

    def setUp(self):
        super(FakeGroupsTest, self).setUp()
        self.network = self.useFixture(FakeNetwork())

    def test_get(self):
        self.network.get('http://test.com', text='data')
        response = requests.get('http://test.com')
        self.assertEqual("data", response.text)



import pwd

from fixtures import Fixture

from ._overlay import Overlay


class FakeUsers(Fixture):

    def _setUp(self):
        self._users = {}
        self.useFixture(Overlay("pwd.getpwnam", self._getpwnam, self._is_fake))

    def add(self, name, uid):
        self._users[name] = (
            "x", uid, uid, name, "/home/{}".format(name), "/bin/bash")

    def _getpwnam(self, real, name):
        return pwd.struct_passwd((name,) + self._users[name])

    def _is_fake(self, name):
        return name in self._users



import glob
import os
import sqlite3
import six
import shutil

from testtools import TestCase, skipIf
from testtools.matchers import (
    DirExists,
    HasPermissions,
    FileContains,
)

from fixtures import TempDir

from ..filesystem import FakeFilesystem
from ..matchers import HasOwnership


if six.PY2:
    PermissionError = OSError


class FakeFilesystemTest(TestCase):

    def setUp(self):
        super(FakeFilesystemTest, self).setUp()
        self.fs = self.useFixture(FakeFilesystem())

    def test_add(self):
        self.fs.add("/foo/bar")
        os.makedirs("/foo/bar")
        self.assertThat("/foo/bar", DirExists())

    def test_add_sub_paths(self):
        self.fs.add("/foo")
        os.makedirs("/foo/bar")
        self.assertThat("/foo/bar", DirExists())

    def test_add_non_absolute(self):
        self.assertRaises(ValueError, self.fs.add, "foo/bar")

    def test_fchown(self):
        self.fs.add("/foo/bar")
        os.mkdir("/foo")
        with open("/foo/bar", "w") as fd:
            os.fchown(fd.fileno(), 123, 456)
        self.assertThat("/foo/bar", HasOwnership(123, 456))

    @skipIf(os.getuid() == 0, "Can't run as root")
    def test_fchown_real(self):
        temp_dir = self.useFixture(TempDir())
        path = temp_dir.join("foo")
        with open(path, "w") as fd:
            self.assertRaises(
                PermissionError, os.fchown, fd.fileno(), 12345, 9999)

    def test_chown(self):
        self.fs.add("/foo/bar")
        os.makedirs("/foo/bar")
        os.chown("/foo/bar", 123, 456)
        self.assertThat("/foo/bar", HasOwnership(123, 456))

    @skipIf(os.getuid() == 0, "Can't run as root")
    def test_chown_real(self):
        temp_dir = self.useFixture(TempDir())
        path = temp_dir.join("foo")
        os.makedirs(path)
        self.assertRaises(PermissionError, os.chown, path, 12345, 9999)

    def test_chmod(self):
        self.fs.add("/foo")
        with open("/foo", "w") as fd:
            fd.write("")
        os.chmod("/foo", 0o600)
        self.assertThat("/foo", HasPermissions("0600"))

    def test_symlink(self):
        self.fs.add("/foo")
        os.mkdir("/foo")
        with open("/foo/bar", "w") as fd:
            fd.write("hello")
        os.symlink("/foo/bar", "/foo/egg")
        self.assertThat("/foo/egg", FileContains("hello"))

    def test_unlink(self):
        self.fs.add("/foo")
        os.mkdir("/foo")
        with open("/foo/bar", "w") as fd:
            fd.write("hello")
        os.unlink("/foo/bar")
        self.assertEqual([], os.listdir("/foo"))

    def test_rmtree(self):
        self.fs.add("/foo")
        os.makedirs("/foo/bar/egg")
        shutil.rmtree("/foo/bar")
        self.assertEqual([], os.listdir("/foo"))

    if six.PY3:

        def test_listdir_with_fd(self):
            self.fs.add("/foo")
            os.makedirs("/foo/bar")
            fd = os.open("/foo", os.O_RDONLY)
            self.addCleanup(os.close, fd)
            self.assertEqual(["bar"], os.listdir(fd))

    def test_readlink_to_real_path(self):
        self.fs.add("/foo")
        os.mkdir("/foo")
        temp_dir = self.useFixture(TempDir())
        os.symlink(temp_dir.path, "/foo/bar")
        self.assertEqual(temp_dir.path, os.readlink("/foo/bar"))

    def test_readlink_to_fake_path(self):
        self.fs.add("/foo")
        os.mkdir("/foo")
        os.symlink("/foo/bar", "/foo/egg")
        self.assertEqual("/foo/bar", os.readlink("/foo/egg"))

    def test_rename(self):
        self.fs.add("/foo")
        os.makedirs("/foo/bar")
        os.rename("/foo/bar", "/foo/egg")
        self.assertThat("/foo/egg", DirExists())

    def test_glob(self):
        self.fs.add("/foo")
        os.makedirs("/foo/bar")
        os.makedirs("/foo/baz")
        self.assertEqual(
            [os.path.join(self.fs.root.path, "foo/bar"),
             os.path.join(self.fs.root.path, "foo/baz")],
            sorted(glob.glob("/foo/ba*")))

    def test_walk(self):
        self.fs.add("/foo")
        os.makedirs("/foo/bar")
        self.assertEqual(
            [(os.path.join(self.fs.root.path, "foo/bar"), [], [])],
            list(os.walk("/foo/bar")))

    def test_sqlite3(self):
        self.fs.add("/foo")
        os.makedirs("/foo/bar")
        conn = sqlite3.connect("/foo/bar/db")
        self.addCleanup(conn.close)
        self.assertIsNotNone(conn)



import os
import six

from fixtures import (
    Fixture,
    TempDir,
)

from ._overlay import Overlay

if six.PY2:
    BUILTIN_OPEN = "__builtin__.open"
if six.PY3:
    BUILTIN_OPEN = "builtins.open"


GENERIC_APIS = (
    BUILTIN_OPEN,
    "glob.glob",
    "os.mkdir",
    "os.rmdir",
    "os.chmod",
    "os.remove",
    "os.unlink",
    "os.listdir",
    "os.walk",
    "os.open",
    "os.path.exists",
    "os.path.isdir",
    "os.utime",
    "sqlite3.connect",
)


class FakeFilesystem(Fixture):
    """A transparent overlay test filesystem tree.

    It allows to transparently redirect filesystem APIs to a temporary
    directory, instead of the actual filesystem.
    """

    def _setUp(self):
        self.root = self.useFixture(TempDir())
        self._paths = {}
        self._ownership = {}

        if six.PY2:
            # Python 2 doesn't support a fd argument
            condition = self._is_fake_path
        if six.PY3:
            condition = self._is_fake_path_or_fd

        for api in GENERIC_APIS:
            self.useFixture(Overlay(api, self._generic, condition))

        self.useFixture(Overlay("os.fchown", self._fchown, self._is_fake_fd))
        self.useFixture(Overlay("os.chown", self._chown, self._is_fake_path))
        self.useFixture(Overlay("os.stat", self._stat, self._is_fake_path))
        self.useFixture(Overlay("os.lstat", self._stat, self._is_fake_path))
        self.useFixture(Overlay(
            "os.symlink", self._symlink, self._is_fake_symlink))
        self.useFixture(
            Overlay("os.readlink", self._readlink, self._is_fake_path))
        self.useFixture(Overlay("os.rename", self._rename, lambda *args: True))

    def add(self, path):
        """Add a path to the overlay filesytem.

        Any filesystem operation involving the this path or any sub-paths
        of it will be transparently redirected to temporary root dir.

        @path: An absolute path string.
        """
        if not path.startswith(os.sep):
            raise ValueError("Non-absolute path '{}'".format(path))
        path = path.rstrip(os.sep)
        while True:
            self._paths[path] = None
            path, _ = os.path.split(path)
            if path == os.sep:
                break

    def _fchown(self, real, fileno, uid, gid):
        """Run fake fchown code if fileno points to a sub-path of our tree.

        The ownership set with this fake fchown can be inspected by looking
        at the self.uid/self.gid dictionaries.
        """
        path = self._fake_path(self._path_from_fd(fileno))
        self._chown_common(path, uid, gid)

    def _chown(self, real, path, uid, gid):
        self._chown_common(path, uid, gid)

    def _chown_common(self, path, uid, gid):
        self._ownership[path] = (uid, gid)

    def _stat(self, real, path, *args, **kwargs):
        info = list(real(self._real_path(path)))
        if path in self._ownership:
            info[4:5] = self._ownership[path]
        return os.stat_result(info)

    def _symlink(self, real, src, dst, *args, **kwargs):
        if self._is_fake_path(src):
            src = self._real_path(src)
        if self._is_fake_path(dst):
            dst = self._real_path(dst)
        return real(src, dst, *args, **kwargs)

    def _readlink(self, real, path, *args, **kwargs):
        result = real(self._real_path(path), *args, **kwargs)
        if result.startswith(self.root.path):
            result = self._fake_path(result)
        return result

    def _rename(self, real, src, dst):
        if self._is_fake_path(src):
            src = self._real_path(src)
        if self._is_fake_path(dst):
            dst = self._real_path(dst)
        return real(src, dst)

    def _generic(self, real, path, *args, **kwargs):
        return real(self._real_path(path), *args, **kwargs)

    def _is_fake_path(self, path, *args, **kwargs):
        for prefix in self._paths:
            if path.startswith(prefix):
                return True
        return False

    def _is_fake_fd(self, fileno, *args, **kwargs):
        path = self._path_from_fd(fileno)
        return path.startswith(self.root.path)

    if six.PY3:

        def _is_fake_path_or_fd(self, path, *args, **kwargs):
            if isinstance(path, int):
                path = self._path_from_fd(path)
            return self._is_fake_path(path)

    def _is_fake_symlink(self, src, dst, *args, **kwargs):
        return self._is_fake_path(src) or self._is_fake_path(dst)

    def _path_from_fd(self, fileno):
        return os.readlink("/proc/self/fd/{}".format(fileno))

    def _real_path(self, path):
        return self.root.join(path.lstrip(os.sep))

    def _fake_path(self, path):
        return path[len(self.root.path):]



import grp

from testtools import TestCase

from ..groups import FakeGroups


class FakeGroupsTest(TestCase):

    def setUp(self):
        super(FakeGroupsTest, self).setUp()
        self.groups = self.useFixture(FakeGroups())

    def test_real(self):
        info = grp.getgrnam("root")
        self.assertEqual(0, info.gr_gid)

    def test_fake(self):
        self.groups.add("foo", 123)
        info = grp.getgrnam("foo")
        self.assertEqual(123, info.gr_gid)



import os
import time
import errno
import socket
import logging
import subprocess

from fixtures import FakeLogger

from testtools import TestCase
from testtools.content import text_content
from testtools.matchers import (
    Contains,
    FileContains,
)

from ..executable import FakeExecutable


class FakeExecutableTest(TestCase):

    def setUp(self):
        super(FakeExecutableTest, self).setUp()
        self.executable = self.useFixture(FakeExecutable())

    def test_out(self):
        self.executable.out("hello")
        self.assertEqual(
            b"hello\n", subprocess.check_output([self.executable.path]))

    def test_sleep(self):
        self.executable.sleep(1)
        self.assertThat(
            self.executable.path,
            FileContains(matcher=Contains("time.sleep(1)")))

    def test_listen_random(self):
        self.executable.listen()
        self.assertIsNotNone(self.executable.port)

    def test_listen(self):
        logger = self.useFixture(FakeLogger(format="%(asctime)s %(message)s"))
        self.executable.listen(6666)
        self.executable.sleep(1)
        self.executable.spawn()

        # Try to connect to the socket
        sock = socket.socket()
        transient = (errno.ECONNREFUSED, errno.ECONNRESET)

        attempts = 10
        for i in range(1, attempts + 1):
            try:
                sock.connect(("127.0.0.1", self.executable.port))
            except socket.error as error:  # pragma: no cover
                if error.errno in transient and i != attempts:
                    time.sleep(0.05 * i)
                    continue
                logging.error("connection attempt %d failed", i)
                raise error
            break
        self.assertEqual("127.0.0.1", sock.getsockname()[0])

    def test_hang(self):
        self.executable.out("hello")  # Used to ensure the process is running
        self.executable.hang()
        process = subprocess.Popen(
            [self.executable.path], stdout=subprocess.PIPE)
        self.assertEqual(b"hello\n", process.stdout.read(6))
        process.terminate()
        self.addCleanup(process.wait)
        self.addCleanup(process.kill)
        self.assertEqual((0, 0), os.waitpid(process.pid, os.WNOHANG))



import io
import os
import argparse


class Wget(object):

    name = "wget"

    def __init__(self, locations=None):
        self.locations = locations or {}

    def __call__(self, proc_args):
        cwd = proc_args.get("cwd") or ""
        parser = argparse.ArgumentParser()
        parser.add_argument("url")
        parser.add_argument("-O", dest="output")
        parser.add_argument("-q", dest="quiet", action="store_true")
        parser.add_argument("-N", dest="timestamping", action="store_true")
        parser.add_argument("--no-check-certificate", action="store_true")
        args = parser.parse_args(proc_args["args"][1:])
        content = self.locations[args.url]
        file = args.url.split("/")[-1]
        file_dir = os.path.join(cwd, file)
        result = {}
        if args.output == "-":
            result["stdout"] = io.BytesIO(content)
        elif (args.output is None):
            with open(file_dir, "wb") as fd:
                fd.write(content)
        else:
            with open(args.output, "wb") as fd:
                fd.write(content)

        return result



from testtools import TestCase

from ..systemctl import Systemctl


class SystemctlTest(TestCase):

    def setUp(self):
        super(SystemctlTest, self).setUp()
        self.systemctl = Systemctl()

    def test_stop(self):
        self.systemctl({"args": ["systemctl", "stop", "foo"]})
        self.assertEqual(["stop"], self.systemctl.actions["foo"])

    def test_start(self):
        self.systemctl({"args": ["systemctl", "start", "foo"]})
        self.assertEqual(["start"], self.systemctl.actions["foo"])

    def test_is_active(self):
        self.systemctl({"args": ["systemctl", "start", "foo"]})
        result = self.systemctl({"args": ["systemctl", "is-active", "foo"]})
        self.assertEqual(0, result.get("returncode"))

    def test_is_not_active(self):
        result = self.systemctl({"args": ["systemctl", "is-active", "foo"]})
        self.assertEqual(3, result["returncode"])



import os

import argparse


class Dpkg(object):

    name = "dpkg"

    def __init__(self):
        self.actions = {}

    def __call__(self, proc_args):
        parser = argparse.ArgumentParser()
        parser.add_argument("-i", dest="install")
        args = parser.parse_args(proc_args["args"][1:])
        if args.install:
            package = os.path.basename(args.install)
            if package.endswith(".deb"):
                package = package[:-len(".deb")]
            package = package.split("_")[0]
            actions = self.actions.setdefault(package, [])
            actions.append("install")
        return {}



from fixtures import Fixture

from requests_mock import Mocker


class FakeNetwork(Fixture):
    """Fixture adapter around request-mock."""

    def _setUp(self):
        self._requests = Mocker()
        self._requests.start()
        self.addCleanup(self._requests.stop)

    def __getattr__(self, name):
        return getattr(self._requests, name)



import os

from pbr.version import VersionInfo

from .users import FakeUsers
from .groups import FakeGroups
from .filesystem import FakeFilesystem
from .processes import FakeProcesses
from .network import FakeNetwork
from .time import FakeTime
from .threads import FakeThreads
from .executable import FakeExecutable


__all__ = [
    "FakeUsers",
    "FakeGroups",
    "FakeFilesystem",
    "FakeProcesses",
    "FakeNetwork",
    "FakeTime",
    "FakeThreads",
    "FakeExecutable",
]


_v = VersionInfo("systemfixtures").semantic_version()
__version__ = _v.release_string()
version_info = _v.version_tuple()


def load_tests(loader, standard_tests, pattern):
    this_dir = os.path.dirname(__file__)
    package_tests = loader.discover(start_dir=this_dir, pattern=pattern)
    standard_tests.addTests(package_tests)
    return standard_tests



import grp

from fixtures import Fixture

from ._overlay import Overlay


class FakeGroups(Fixture):

    def _setUp(self):
        self._groups = {}
        self.useFixture(Overlay("grp.getgrnam", self._getgrnam, self._is_fake))

    def add(self, name, gid):
        self._groups[name] = ("x", gid, [])

    def _getgrnam(self, real, name):
        return grp.struct_group((name,) + self._groups[name])

    def _is_fake(self, name):
        return name in self._groups



import os
import socket

import six

from testtools.content import Content
from testtools.content_type import UTF8_TEXT

from fixtures import (
    Fixture,
    TempDir,
)

if six.PY2:
    import subprocess32 as subprocess
if six.PY3:
    import subprocess


class FakeExecutable(Fixture):
    """Create Python scripts that mimic the behavior of real executables."""

    def _setUp(self):
        self.path = self.useFixture(TempDir()).join("executable")
        self.line("#!/usr/bin/env python")
        self.line("import logging")
        self.line("logging.basicConfig("
                      "format='%(asctime)s %(message)s', level=logging.DEBUG)")
        os.chmod(self.path, 0o0755)

        self._process = None
        self.addDetail("fake-process", Content(UTF8_TEXT, self._process_info))

    def spawn(self):
        """Spawn the fake executable using subprocess.Popen."""
        self._process = subprocess.Popen(
            [self.path], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        self.addCleanup(self._process_kill)

    def out(self, text):
        self.line("import sys")
        self.line("sys.stdout.write('{}\\n')".format(text))
        self.line("sys.stdout.flush()")

    def log(self, message):
        self.line("logging.info('{}')".format(message))

    def sleep(self, seconds):
        self.line("import time")
        self.line("time.sleep({})".format(seconds))

    def hang(self):
        self.line("import time")
        self.line("import signal")
        self.line("signal.signal(signal.SIGTERM, lambda *args: None)")
        self.out("hanging")
        self.line("while True: time.sleep(1)")

    def listen(self, port=None):
        """Make the fake executable listen to the specified port.

        Possible values for 'port' are:

        - None: Allocate immediately a free port and instruct the fake
          executable to use it when it's invoked. This is subject to
          a race condition, if that port that was free when listen() was
          invoked later becomes used before the fake executable had chance
          to bind to it. However it has the advantage of exposing the
          free port as FakeExecutable.port instance variable, that can easily
          be consumed by tests.

        - An integer: Listen to this specific port.
        """
        if port is None:
            port = allocate_port()
        self.port = port
        self.line("import socket")
        self.line("sock = socket.socket()")
        self.line("sock.bind(('localhost', {}))".format(self.port))
        self.log("listening: %d" % self.port)
        self.line("sock.listen(0)")

    def line(self, line):
        with open(self.path, "a") as fd:
            fd.write("{}\n".format(line))

    def _process_kill(self):
        """Kill the fake executable process if it's still running."""
        if self._process.poll() is None:  # pragma: no cover
            self._process.kill()
            self._process.wait(timeout=5)

    def _process_info(self):
        """Return details about the fake process."""
        if not self._process:
            return []

        output, error = self._process.communicate(timeout=5)
        if error is None:
            error = b""
        output = output.decode("utf-8").strip()
        error = error.decode("utf-8").strip()
        info = (u"returncode: %r\n"
                u"output:\n%s\n"
                u"error:\n%s\n" % (self._process.returncode, output, error))
        return [info.encode("utf-8")]


def get_port(socket):
    """Return the port to which a socket is bound."""
    addr, port = socket.getsockname()
    return port


def allocate_port():
    """Allocate an unused port.

    There is a small race condition here (between the time we allocate the
    port, and the time it actually gets used), but for the purposes for which
    this function gets used it isn't a problem in practice.
    """
    sock = socket.socket()
    try:
        sock.bind(("localhost", 0))
        return get_port(sock)
    finally:
        sock.close()
